package v3d.editor;

import java.applet.*;
import java.awt.*;
import java.io.*;
import java.util.*;

/*
 * This is the stand-alone version of FileIO class. ioPanel now has four buttons:
 * "Load","Save","Export VRML","Export Image".
 */
class ApplicationFileIO extends FileIO {
	public ApplicationFileIO(Editor editor) {
		super(editor);
		ioPanel = new ApplicationFileIOPanel(this);
	}

	// starts the FileIO from editor start function
	public void start() {
		Date d = new Date();
		editor.setCurrentEditor("Editing session " + d.toLocaleString());
		editor.setUserMsg("Messages not available in the standalone version.");
	}

	public void load(String directoryName,String fileName) {
		if (fileName != null) {
			Scene scene = editor.getScene();

			File file = new File(directoryName,fileName);
			FileInputStream is = null;
			try {
				is = new FileInputStream(file);
				scene.deserialize(is);
			}
			catch(Exception e) {
				scene.reset();
//				System.out.println("Error: "+e.toString());
				QuerySpawner query = new QuerySpawner();
				query.doQuery("",e.toString(),"Ok",null,null);
			}

			try {
				if (is != null)
					is.close();
			}
			catch(Exception e) {
			}

			Date d = new Date();
			editor.notifySceneLoaded();
		}
	}

	public String loadString(String directoryName,String fileName) {
		if (fileName != null) {
			String s = new String();

			FileInputStream is = null;
			try {
				is = new FileInputStream(fileName);
				DataInputStream dis = new DataInputStream(is);
				// maybe we should rather use Unicode than 8-bit charset
				while (dis.available() > 0)
					s += dis.readLine();
			}
			catch(Exception e) {
//				System.out.println("Error: "+e.toString());
				QuerySpawner query = new QuerySpawner();
				query.doQuery("",e.toString(),"Ok",null,null);
				s = null;
			}

			try {
				if (is != null)
				is.close();
			}
			catch(Exception e) {
			}

			return s;
		}
		return null;
	}

	public void save(String directoryName,String fileName) {
		Scene sc = editor.getScene();

		// Fetch the scene in native 3d6b format. Currently the format is
		// string-based, but it may change to binary to save bandwith and space
		String scene;
		scene = sc.serialize();

		File file = new File(directoryName,fileName);
		// Output the buffer somewhere
		try {
			FileOutputStream out = new FileOutputStream(file);
			DataOutputStream dout = new DataOutputStream(out);
			dout.writeBytes(scene);
		}
		catch(Exception e) {
//			System.out.println("Error: "+e.toString());
			QuerySpawner query = new QuerySpawner();
			query.doQuery("",e.toString(),"Ok",null,null);
		}
	}

	public void exportVRML(String directoryName,String fileName) {
		Scene sc = editor.getScene();

		// Fetch the scene in VRML format.
		String vrml;
		vrml = sc.getVRML();

		File file = new File(directoryName,fileName);
		// Output the buffer somewhere
		try {
			FileOutputStream out = new FileOutputStream(file);
			DataOutputStream dout = new DataOutputStream(out);
			dout.writeBytes(vrml);
		}
		catch(Exception e) {
//			System.out.println("Error: "+e.toString());
			QuerySpawner query = new QuerySpawner();
			query.doQuery("",e.toString(),"Ok",null,null);
		}
	}

	public void exportImage(String directoryName,String fileName) {
		Scene sc = editor.getScene();

		// Fetch the image of the scene in Targa format.
		// The image can be converted to jpeg on the server side.
		byte tga[];
		tga = editor.getThumbnailTGA(800,400,1);
		File file = new File(directoryName,fileName);
		// Output the buffer
		try {
			FileOutputStream out = new FileOutputStream(file);
			out.write(tga);
		}
		catch(Exception e) {
//			System.out.println("Error: "+e.toString());
			QuerySpawner query = new QuerySpawner();
			query.doQuery("",e.toString(),"Ok",null,null);
		}
	}
}

/*
 * This is the UI panel for the FileIO. It is incorporated to the rest of the
 * user interface, but it is isolated here so that it stays as its own module
 * and can be easily changed.
 */
class ApplicationFileIOPanel extends Panel {
	ApplicationFileIO io;

	ApplicationFileIOPanel(ApplicationFileIO io)
	{
		this.io = io;
		this.setLayout(new GridLayout(4,1,2,2));
		this.add(new Button("Load"));
		this.add(new Button("Save"));
		this.add(new Button("Export VRML"));
		this.add(new Button("Export Image"));
	}

	public boolean action(Event e, Object arg) {
		if (e.target instanceof Button) {
			String buttonText = (String)arg;
			if (buttonText.equals("Load")) {
				Frame tf = new Frame();
				FileDialog dialog = new FileDialog(tf,"Load a file",FileDialog.LOAD);
				dialog.show();
				String fname = dialog.getFile();
				String dir = dialog.getDirectory();
				if (fname != null && dir != null)
					io.load(dir,fname);
				tf.dispose();
			}
			else if (buttonText.equals("Save")) {
				Frame tf = new Frame();
				FileDialog dialog = new FileDialog(tf,"Save a file",FileDialog.SAVE);
				dialog.show();
				String fname = dialog.getFile();
				String dir = dialog.getDirectory();
				if (fname != null && dir != null)
					io.save(dir,fname);
				tf.dispose();
			}
			else if (buttonText.equals("Export VRML")) {
				Frame tf = new Frame();
				FileDialog dialog = new FileDialog(tf,"Export VRML",FileDialog.SAVE);
				dialog.show();
				String fname = dialog.getFile();
				String dir = dialog.getDirectory();
				if (fname != null && dir != null)
					io.exportVRML(dir,fname);
				tf.dispose();
			}
			else if (buttonText.equals("Export Image")) {
				Frame tf = new Frame();
				FileDialog dialog = new FileDialog(tf,"Export TGA image",FileDialog.SAVE);
				dialog.show();
				String fname = dialog.getFile();
				String dir = dialog.getDirectory();
				if (fname != null && dir != null)
					io.exportImage(dir,fname);
				tf.dispose();
			}
		}
		return true;
	}
}
