/*
 * This is a class that implements a 4x4 matrix and various
 * operations related to it.
 *
 */
package v3d.editor;

class Matrix4x4
{
	public float values[][];
	
	Matrix4x4() {
		this.values = new float[4][4];
	}
	
	Matrix4x4(Matrix4x4 matrix) {
		this.values = new float[4][4];
		int i,j;
		for (i=0; i<4; i++) {
			for (j=0; j<4; j++) {
				this.values[i][j] = matrix.values[i][j];
			}
		}
	}
	
	Matrix4x4(float matrix[][]) {
		this.values = new float[4][4];
		int i,j;
		for (i=0; i<4; i++) {
			for (j=0; j<4; j++) {
				this.values[i][j] = matrix[i][j];
			}
		}
	}

	public void copy(Matrix4x4 matrix) {
		int i,j;
		for (i=0; i<4; i++) {
			for (j=0; j<4; j++) {
				this.values[i][j] = matrix.values[i][j];
			}
		}
	}

	public Matrix4x4 multiply(Matrix4x4 m) {
		Matrix4x4 tMatrix = new Matrix4x4();

		int i, j;
		for (i = 0; i < 4; i++) {
			for (j = 0; j < 4; j++) {
				tMatrix.values[j][i] =
					this.values[j][0] * m.values[0][i] +
					this.values[j][1] * m.values[1][i] +
					this.values[j][2] * m.values[2][i] +
					this.values[j][3] * m.values[3][i];
			}
		}

		return tMatrix;
	}

	public Matrix4x4 multiply(float f) {
		Matrix4x4 tMatrix = new Matrix4x4();

		int i, j;
		for (i = 0; i < 4; i++) {
			for (j = 0; j < 4; j++) {
				tMatrix.values[i][j] = this.values[i][j] * f;
			}
		}

		return tMatrix;
	}

	public void makeIdentity() {
		this.values[0][0] = 1f;
		this.values[0][1] = 0f;
		this.values[0][2] = 0f;
		this.values[0][3] = 0f;

		this.values[1][0] = 0f;
		this.values[1][1] = 1f;
		this.values[1][2] = 0f;
		this.values[1][3] = 0f;

		this.values[2][0] = 0f;
		this.values[2][1] = 0f;
		this.values[2][2] = 1f;
		this.values[2][3] = 0f;

		this.values[3][0] = 0f;
		this.values[3][1] = 0f;
		this.values[3][2] = 0f;
		this.values[3][3] = 1f;
	}

	public void makeRotationX(float angle) {
		float sinAngle = (float)Math.sin(angle);
		float cosAngle = (float)Math.cos(angle);

		this.values[0][0] = 1f;
		this.values[0][1] = 0f;
		this.values[0][2] = 0f;
		this.values[0][3] = 0f;

		this.values[1][0] = 0f;
		this.values[1][1] = cosAngle;
		this.values[1][2] = sinAngle;
		this.values[1][3] = 0f;

		this.values[2][0] = 0f;
		this.values[2][1] = -sinAngle;
		this.values[2][2] = cosAngle;
		this.values[2][3] = 0f;

		this.values[3][0] = 0f;
		this.values[3][1] = 0f;
		this.values[3][2] = 0f;
		this.values[3][3] = 1f;
	}

	public void makeRotationY(float angle) {
		float sinAngle = (float)Math.sin(angle);
		float cosAngle = (float)Math.cos(angle);

		this.values[0][0] = cosAngle;
		this.values[0][1] = 0f;
		this.values[0][2] = -sinAngle;
		this.values[0][3] = 0f;

		this.values[1][0] = 0f;
		this.values[1][1] = 1f;
		this.values[1][2] = 0f;
		this.values[1][3] = 0f;

		this.values[2][0] = sinAngle;
		this.values[2][1] = 0f;
		this.values[2][2] = cosAngle;
		this.values[2][3] = 0f;

		this.values[3][0] = 0f;
		this.values[3][1] = 0f;
		this.values[3][2] = 0f;
		this.values[3][3] = 1f;
	}

	public void makeRotationZ(float angle) {
		float sinAngle = (float)Math.sin(angle);
		float cosAngle = (float)Math.cos(angle);

		this.values[0][0] = cosAngle;
		this.values[0][1] = sinAngle;
		this.values[0][2] = 0f;
		this.values[0][3] = 0f;

		this.values[1][0] = -sinAngle;
		this.values[1][1] = cosAngle;
		this.values[1][2] = 0f;
		this.values[1][3] = 0f;

		this.values[2][0] = 0f;
		this.values[2][1] = 0f;
		this.values[2][2] = 1f;
		this.values[2][3] = 0f;

		this.values[3][0] = 0f;
		this.values[3][1] = 0f;
		this.values[3][2] = 0f;
		this.values[3][3] = 1f;
	}

	public void makeTranslation(Vector3D offset) {
		this.values[0][0] = 1f;
		this.values[0][1] = 0f;
		this.values[0][2] = 0f;
		this.values[0][3] = 0f;

		this.values[1][0] = 0f;
		this.values[1][1] = 1f;
		this.values[1][2] = 0f;
		this.values[1][3] = 0f;

		this.values[2][0] = 0f;
		this.values[2][1] = 0f;
		this.values[2][2] = 1f;
		this.values[2][3] = 0f;

		this.values[3][0] = offset.x;
		this.values[3][1] = offset.y;
		this.values[3][2] = offset.z;
		this.values[3][3] = 1f;
	}

	public void rotateX(float angle) {
		Matrix4x4 rotation = new Matrix4x4();
		rotation.makeRotationX(angle);
		this.copy(rotation.multiply(this));
	}

	public void rotateY(float angle) {
		Matrix4x4 rotation = new Matrix4x4();
		rotation.makeRotationY(angle);
		this.copy(rotation.multiply(this));
	}

	public void rotateZ(float angle) {
		Matrix4x4 rotation = new Matrix4x4();
		rotation.makeRotationZ(angle);
		this.copy(rotation.multiply(this));
	}

	public void translate(Vector3D offset) {
		Matrix4x4 translation = new Matrix4x4();
		translation.makeTranslation(offset);
		this.copy(translation.multiply(this));
	}

	public void scale(float scale) {
		this.values[0][0] *= scale;
		this.values[1][1] *= scale;
		this.values[2][2] *= scale;
	}

	public void transform(Vector3D src, Vector3D dst) {
		dst.x = this.values[0][0] * src.x +
			this.values[1][0] * src.y +
			this.values[2][0] * src.z + this.values[3][0];
		dst.y = this.values[0][1] * src.x +
			this.values[1][1] * src.y +
			this.values[2][1] * src.z + this.values[3][1];
		dst.z = this.values[0][2] * src.x +
			this.values[1][2] * src.y +
			this.values[2][2] * src.z + this.values[3][2];
	}


	public void transpose() {
		Matrix4x4 tMatrix = new Matrix4x4(this);
		int i,j;

		for (i=0; i<4; i++) {
			for (j=0; j<4; j++) {
				this.values[i][j] = tMatrix.values[j][i];
			}
		}
	}

	public void inverseFast() {
		Matrix4x4 tMatrix = new Matrix4x4(this);
		this.values[0][0] = tMatrix.values[0][0];
		this.values[0][1] = tMatrix.values[1][0];
		this.values[0][2] = tMatrix.values[2][0];
		this.values[1][0] = tMatrix.values[0][1];
		this.values[1][1] = tMatrix.values[1][1];
		this.values[1][2] = tMatrix.values[2][1];
		this.values[2][0] = tMatrix.values[0][2];
		this.values[2][1] = tMatrix.values[1][2];
		this.values[2][2] = tMatrix.values[2][2];
	
		this.values[0][3] = 0;
		this.values[1][3] = 0;
		this.values[2][3] = 0;
		this.values[3][3] = 1;

		Vector3D pos = new Vector3D(tMatrix.values[3][0], tMatrix.values[3][1], tMatrix.values[3][2]);
		Vector3D ox = new Vector3D(tMatrix.values[0][0], tMatrix.values[0][1], tMatrix.values[0][2]);
		Vector3D oy = new Vector3D(tMatrix.values[1][0], tMatrix.values[1][1], tMatrix.values[1][2]);
		Vector3D oz = new Vector3D(tMatrix.values[2][0], tMatrix.values[2][1], tMatrix.values[2][2]);

		this.values[3][0] = -pos.dotProd(ox);
		this.values[3][1] = -pos.dotProd(oy);
		this.values[3][2] = -pos.dotProd(oz);
	}

	public void normalize() {
		Vector3D x = new Vector3D(this.values[0][0], this.values[1][0], this.values[2][0]);
		Vector3D z = new Vector3D(this.values[0][2], this.values[1][2], this.values[2][2]);

		Vector3D newy = z.crossProd(x);
		Vector3D newz = x.crossProd(newy);
		x.normalizeThis();
		newy.normalizeThis();
		newz.normalizeThis();

		this.values[0][0] = x.x;
		this.values[1][0] = x.y;
		this.values[2][0] = x.z;
		this.values[0][1] = newy.x;
		this.values[1][1] = newy.y;
		this.values[2][1] = newy.z;
		this.values[0][2] = newz.x;
		this.values[1][2] = newz.y;
		this.values[2][2] = newz.z;
	}

	public void makeLookAt(Vector3D pos, Vector3D at, Vector3D up) {
/*		System.out.println("pos:");
		pos.print();
		System.out.println("at:");
		at.print();
		System.out.println("up:");
		up.print();
*/

		Vector3D viewVec = at.subtract(pos);
 
//		System.out.println("viewVec:");
//		viewVec.print();

		float length = viewVec.length(); 
		viewVec.divideThis(length);
 
		float dot = viewVec.dotProd(up);
		Vector3D upVec = up.subtract(viewVec.multiply(dot));

		upVec.normalizeThis();

//		System.out.println("upVec:");
//		upVec.print();

		Vector3D rightVec = upVec.crossProd(viewVec);

//		System.out.println("rightVec:");
//		rightVec.print();

		this.values[0][0] = rightVec.x;
		this.values[1][0] = upVec.x;
		this.values[2][0] = viewVec.x;
		this.values[0][1] = rightVec.y;
		this.values[1][1] = upVec.y;
		this.values[2][1] = viewVec.y;
		this.values[0][2] = rightVec.z;
		this.values[1][2] = upVec.z;
		this.values[2][2] = viewVec.z;
 
		this.values[3][0] = pos.x;
		this.values[3][1] = pos.y;
		this.values[3][2] = pos.z;

		this.values[0][3] = 0;
		this.values[1][3] = 0;
		this.values[2][3] = 0;
		this.values[3][3] = 1;
	}

	public void makeProjection(float width, float height, float fov, float zNear, float zFar) {
		float aspect = width / height;

		float fovRatio = (float)(Math.cos(fov / 2) / Math.sin(fov / 2));

		float w = fovRatio;
		float h = aspect * fovRatio;
		float q = zFar / (zFar - zNear);

		this.values[0][0] = w;
		this.values[0][1] = 0;
		this.values[0][2] = 0;
		this.values[0][3] = 0;

		this.values[1][0] = 0;
		this.values[1][1] = h;
		this.values[1][2] = 0;
		this.values[1][3] = 0;

		this.values[2][0] = 0;
		this.values[2][1] = 0;
		this.values[2][2] = q;
		this.values[2][3] = 1.0f;

		this.values[3][0] = 0;
		this.values[3][1] = 0;
		this.values[3][2] = -q * zNear;
		this.values[3][3] = 0;
	}

	public void project(Vector4D src, Vector4D dst) {
		dst.x = src.x * this.values[0][0] + src.y * this.values[1][0] +
			src.z * this.values[2][0] + src.w * this.values[3][0];

		dst.y = src.x * this.values[0][1] + src.y * this.values[1][1] +
			src.z * this.values[2][1] + src.w * this.values[3][1];

		dst.z = src.x * this.values[0][2] + src.y * this.values[1][2] +
			src.z * this.values[2][2] + src.w * this.values[3][2];

		dst.w = src.x * this.values[0][3] + src.y * this.values[1][3] +
			src.z * this.values[2][3] + src.w * this.values[3][3];
	}

	public void print() {
		System.out.println(	" 00: " + String.valueOf(values[0][0]) +
						" 01: " + String.valueOf(values[0][1]) +
						" 02: " + String.valueOf(values[0][2]) +
						" 03: " + String.valueOf(values[0][3]));
		System.out.println(	" 10: " + String.valueOf(values[1][0]) +
						" 11: " + String.valueOf(values[1][1]) +
						" 12: " + String.valueOf(values[1][2]) +
						" 13: " + String.valueOf(values[1][3]));
		System.out.println(	" 20: " + String.valueOf(values[2][0]) +
						" 21: " + String.valueOf(values[2][1]) +
						" 22: " + String.valueOf(values[2][2]) +
						" 23: " + String.valueOf(values[2][3]));
		System.out.println(	" 30: " + String.valueOf(values[3][0]) +
						" 31: " + String.valueOf(values[3][1]) +
						" 32: " + String.valueOf(values[3][2]) +
						" 33: " + String.valueOf(values[3][3]));
	}
}
